from pymacaron.log import pymlogger
from pprint import pformat
from flask import jsonify


log = pymlogger(__name__)


class PyMacaronException(BaseException):
    code = 'UNKNOWN_EXCEPTION'
    status = 500
    error_id = None
    user_message = None
    error_caught = None

    def __str__(self):
        return f'{self.__class__.__name__}({self.status}|{self.code}|{self.user_message})'

    def __init__(self, msg=None):
        self.user_message = msg

    def caught(self, error):
        self.error_caught = error
        return self

    def jsonify(self):
        """Return a Flask reply object describing this error"""
        data = {
            'status': self.status,
            'error': self.code.upper(),
            'error_description': str(self)
        }

        if self.error_caught:
            data['error_caught'] = pformat(self.error_caught)

        if self.error_id:
            data['error_id'] = self.error_id

        if self.user_message:
            data['user_message'] = self.user_message

        r = jsonify(data)
        r.status_code = self.status

        if str(self.status) != "200":
            log.warn("ERROR: caught error %s %s [%s]" % (self.status, self.code, str(self)))

        return r

    def to_dict(self):
        """Return a dict representation of this pymacaron error"""
        d = {
            'status': self.status,
            'error': self.code.upper(),
        }
        if self.error_id:
            d['error_id'] = self.error_id
        if self.user_message:
            d['user_message'] = self.user_message
        if self.error_caught:
            d['error_caught'] = pformat(self.error_caught)
        return d


class InternalValidationError(PyMacaronException):
    code = 'INTERNAL_VALIDATION_ERROR'
    status = 500


class BadResponseException(PyMacaronException):
    code = 'BAD_RESPONSE'
    status = 500


class UnhandledServerError(PyMacaronException):
    code = 'SERVER_ERROR'
    status = 500


class InvalidParameterError(PyMacaronException):
    code = 'INVALID_PARAMETER'
    status = 400


class RequestTimeout(PyMacaronException):
    code = 'REQUEST_TIMEOUT'
    status = 408


#
# Interface to allow creating further Exception classes
#

code_to_class = {}

def add_error(name=None, code=None, status=None):
    """Create a new Exception class"""
    if not name or not status or not code:
        raise Exception("Can't create Exception class %s: you must set both name, status and code" % name)
    myexception = type(name, (PyMacaronException, ), {"code": code, "status": status})
    globals()[name] = myexception
    if code in code_to_class:
        raise Exception("ERROR! Exception %s is already defined." % code)
    code_to_class[code] = myexception
    return myexception


add_error('InternalServerError', 'SERVER_ERROR', 500)
add_error('AuthMissingHeaderError', 'AUTHORIZATION_HEADER_MISSING', 401)
add_error('AuthTokenExpiredError', 'TOKEN_EXPIRED', 401)
add_error('AuthInvalidTokenError', 'TOKEN_INVALID', 401)
# add_error('ValidationError', 'INVALID_PARAMETER', 400)


def is_error(o):
    """True if o is a Flask error Response"""
    if hasattr(o, 'status') and int(o.status) >= 300:
        return True
    return False



import os
import sys
import logging
import click
import pkg_resources
from datetime import datetime
from uuid import uuid4
from flask import Response, redirect, abort
from flask_compress import Compress
from flask_cors import CORS
from pymacaron.apiloader import load_api_models_and_endpoints
from pymacaron.log import set_level, pymlogger
from pymacaron.config import get_config
from pymacaron.monitor import monitor_init
from pymacaron.crash import set_error_reporter
from pymacaron.api import add_ping_hook


log = pymlogger(__name__)


class modelpool():
    """The modelpool of an api is a class whose attributes are all the Pymacaron
    model classes declared in that api
    """

    def __init__(self, name):
        self.api_name = name
        self._model_names = []

    def get_model_names(self):
        return self._model_names

    def add_model(self, model_name, model_class):
        self._model_names.append(model_name)
        setattr(self, model_name, model_class)

    def __getattr__(self, model_name):
        raise Exception(f'Either {self.api_name}.yaml has not been loaded or it does not define object {model_name}')

    def json_to_model(self, model_name, j, keep_datetime=None, prune_none=True):
        """Given a model name and json dict, return an instantiated pymacaron object"""
        # TODO: keep_datetime in this method should be deprecated...
        assert keep_datetime is not False, "Support for keep_datetime=False not implemented"
        return getattr(self, model_name).from_json(j, prune_none=prune_none)


class apipool():
    """The apipool contains the modelpools of all loaded apis"""

    # api_name: api_path
    __api_paths = {}

    @classmethod
    def add_model(cls, api_name, model_name, model_class):
        """Register a model class defined in an api"""
        # Set modelpool.<model_name> to model_class. This allows writing:
        #   from pymacaron import apipool
        #   o = apipool.ping.Ok()
        if not hasattr(apipool, api_name):
            setattr(apipool, api_name, modelpool(api_name))
        models = getattr(apipool, api_name)
        models.add_model(model_name, model_class)

    @classmethod
    def get_model(cls, api_name):
        """Return the pymacaron modelpool for this api"""
        assert hasattr(apipool, api_name), f"Api {api_name} is not loaded in apipool"
        return getattr(apipool, api_name)

    @classmethod
    def get_api_names(cls):
        """Return the names of loaded apis"""
        return sorted(apipool.__api_paths.keys())

    @classmethod
    def load_swagger(cls, api_name, api_path, dest_dir=None, create_endpoints=False, force=False, model_file=None, app_file=None):
        """Load a swagger/openapi specification into pymacaron: generate its model
        classes (declared with pydantic), and optionally generate the Flask api
        endpoints binding endpoint methods to routes.

        Syntax:
            apipool.load_swagger('ping', '../apis/ping.yaml')

        api_name : str
            Name of the api, used to access api models.
        api_path : str
            Path of the swagger file of the api.
        dest_dir: str
            Optional. Path to a directory under which to write the generated
            '<api_name>_models.py' and '<api_name>_app.py' files. Defaults to
            the same directory as the swagger file.
        create_endpoints: bool
            Optional. Set to false to only generate model declarations, and not
            endpoint declarations. Defaults to true.
        force: bool
            Optional. Force regenerating the model and endpoint code even if
            the code files are up to date with the swagger file. Defaults to
            false.

        """

        app_pkg = load_api_models_and_endpoints(
            api_name=api_name,
            api_path=api_path,
            dest_dir=dest_dir,
            create_endpoints=create_endpoints,
            force=force,
            model_file=model_file,
            app_file=app_file,
        )

        # Remember where this api's swagger is located
        apipool.__api_paths[api_name] = api_path

        return app_pkg

    @classmethod
    def enforce_global_model_names(cls):
        """Find all models that have the same names in different apis and compare their
        schema declarations.  Raise an error if at least two of them differ,
        and show all the differences that were found as a diff of their json
        schemas.
        """

        model_name_to_apis = {}

        for api_name in apipool.__api_paths.keys():
            api = getattr(apipool, api_name)
            for model_name in api.get_model_names():
                if model_name not in model_name_to_apis:
                    model_name_to_apis[model_name] = []
                model_name_to_apis[model_name].append(api)

        found_names = []

        for model_name in sorted(model_name_to_apis.keys()):
            apis = model_name_to_apis[model_name]
            cnt = len(apis)
            if cnt > 1:
                log.info(f"Found {model_name} in {cnt} apis: {', '.join([str(a) for a in apis])}")
                m0 = getattr(apis[0], model_name)
                for api in apis[1:]:
                    m1 = getattr(api, model_name)
                    diff = m0.diff_with(m1)
                    if diff:
                        log.error('\n' + diff)
                        if model_name not in found_names:
                            found_names.append(model_name)

        if found_names:
            raise Exception(f"Found models with same names but different schemas in different apis: {', '.join(found_names)}")

    @classmethod
    def publish_apis(cls, app, path='doc', toc=None):
        """Add routes to the Flask app to publish all loaded swagger files under the
        paths doc/<api_name>.yaml and doc/<api_name>. Optionally add a table of
        content to all swagger specs.

        """

        if not apipool.__api_paths:
            raise Exception("You must call .load_apis() before .publish_apis()")

        conf = get_config()

        # Infer the live host url from pym-config.yaml
        proto = 'https'
        live_host = f"{proto}://{conf.live_host}"

        # Allow cross-origin calls
        CORS(app, resources={r"/%s/*" % path: {"origins": "*"}})

        def doc_endpoint(name=None):
            api_name = name.replace('.yaml', '')
            if api_name not in apipool.__api_paths:
                log.error(f"Unknown api name '{api_name}'")
                abort(404)

            api_path = apipool.__api_paths[api_name]

            if name.endswith('.yaml'):
                # Show the swagger file
                with open(api_path, 'r') as f:
                    spec = f.read()

                    if toc:
                        # Insert the table of content after the first 'description: |'
                        spec = spec.replace(
                            'description: |',
                            f'description: |\n{toc}\n',
                            1,
                        )

                    log.info("Returning %s" % api_path)
                    return Response(spec, mimetype='text/plain')

            else:
                # Redirect to swagger-UI at petstore, to open this swagger file
                url = f'http://petstore.swagger.io/?url={live_host}/{path}/{api_name}.yaml'
                log.info(f"Redirecting to {url}")
                return redirect(url, code=302)

        path = path.strip('/')
        route = f'/{path}/<name>'
        log.info(f"Publishing apis under route {route}")
        app.add_url_rule(route, str(uuid4()), view_func=doc_endpoint, methods=['GET'])


class jsonencoders:

    # Default encoder when serializing datetime to Flask response
    __json_encoders = {
        datetime: lambda d: d.isoformat(),
    }

    @classmethod
    def set_json_encoders(cls, encoders):
        assert type(encoders) is dict, "json_encoders must be a dict of type/callable"
        assert datetime in encoders, "json_encoders must define an encoder for datetime"
        assert callable(encoders[datetime]), "json_encoders[datetime] must be a callable"
        cls.__json_encoders = encoders


    @classmethod
    def get_datetime_encoder(cls):
        assert datetime in jsonencoders.__json_encoders
        return jsonencoders.__json_encoders[datetime]


def get_port():
    """Find which TCP port to listen to, based on environment variables"""
    if 'PORT' in os.environ:
        port = os.environ['PORT']
        log.info("Environment variable PORT is set: will listen on port %s" % port)
    elif 'PYM_SERVER_PORT' in os.environ:
        port = os.environ['PYM_SERVER_PORT']
        log.info("Environment variable PYM_SERVER_PORT is set: will listen on port %s" % port)
    else:
        port = 80
        log.info("No HTTP port specified. Will listen on port 80")

    return port



#
# API: class to define then run a micro service api
#

class API(object):


    def __init__(self, app, host='localhost', port=None, debug=False, log_level=logging.DEBUG, json_encoders=None, error_reporter=None, error_callback=None, default_user_id=None, ping_hook=[]):
        """

        Configure the Pymacaron microservice prior to starting it. Arguments:

        app : the flask app

        port : (optional) the http port to listen on (defaults to 80)

        debug : (optional) whether to run with flask's debug mode (defaults to False)

        error_callback : (optional) takes the exception caught while executing an endpoint and return its json representation in the microservice's own error format

        error_reporter : (optional) takes the exception caught in an endpoint together with a title and long text trace, and reports them wherever is relevant

        log_level : (optional) the microservice's log level (defaults to logging.DEBUG)

        ping_hook : (optional) a function to call each time Amazon calls the ping endpoint, which happens every few seconds

        json_encoders: (optional) custom pydantic json encoders, to use when serializing pymacaron models to json

        """
        assert app
        assert port

        self.app = app
        self.port = port
        self.host = host
        self.debug = debug
        self.error_callback = error_callback
        self.error_reporter = error_reporter
        self.ping_hook = ping_hook
        self.app_pkgs = []

        if not port:
            self.port = get_port()

        if default_user_id:
            self.default_user_id = default_user_id

        set_level(log_level)

        if json_encoders:
            log.info(f"Using custom json encoder when serializing Flask response: {json_encoders}")
            jsonencoders.set_json_encoders(json_encoders)

        log.info("Initialized API (%s:%s) (Flask debug:%s)" % (host, port, debug))


    def publish_apis(self, path='doc', toc=None):
        """Publish all loaded apis on under the uri /<path>/<api-name>, by redirecting
        to http://petstore.swagger.io/. Optionally add a common table of
        content (provided in markdown) to all apis' yaml files.
        """
        apipool.publish_apis(self.app, path=path, toc=toc)


    def load_builtin_apis(self, names=['ping']):
        """Load some or all of the builtin apis 'ping' and 'crash'"""
        for name in names:
            yaml_path = pkg_resources.resource_filename(__name__, 'pymacaron/%s.yaml' % name)
            if not os.path.isfile(yaml_path):
                yaml_path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), '%s.yaml' % name)
            app_pkg = apipool.load_swagger(
                name,
                yaml_path,
                dest_dir=get_config().apis_path,
                create_endpoints=True,
            )
            if app_pkg:
                self.app_pkgs.append(app_pkg)


    def load_clients(self, path=None, apis=[]):
        """Generate client libraries for the given apis, without starting an
        api server"""

        if not path:
            raise Exception("Missing path to api swagger files")

        if type(apis) is not list:
            raise Exception("'apis' should be a list of api names")

        if len(apis) == 0:
            raise Exception("'apis' is an empty list - Expected at least one api name")

        for api_name in apis:
            api_path = os.path.join(path, '%s.yaml' % api_name)
            if not os.path.isfile(api_path):
                raise Exception("Cannot find swagger specification at %s" % api_path)
            apipool.load_swagger(
                api_name,
                api_path,
                dest_dir=path,
                create_endpoints=False,
            )

        return self


    def load_apis(self, path=None, ignore=[], only_models=[], force=False, global_names=False):
        """Find all swagger files under the given path. Ignore those whose name is in
        the ignore list. Generate and load models for all others. Generate
        Flask app code for all except those in ignore and only_models list.

        If global_names is true, load_apis() will check all generated models
        from all loaded apis and if two models with the same name but different
        schemas are found, raise an exception and show the diff between those
        two models.

        """

        if not path:
            path = get_config().apis_path

        if type(ignore) is not list:
            raise Exception("'ignore' should be a list of api names")

        if type(only_models) is not list:
            raise Exception("'ignore' should be a list of api names")

        # Always ignore pym-config.yaml
        ignore.append('pym-config')

        # Find all swagger apis under 'path'
        apis = {}

        log.debug("Searching swagger files under path %s" % path)
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.startswith('.#') or f.startswith('#'):
                    log.info("Ignoring file %s" % f)
                elif f.endswith('.yaml'):
                    api_name = f.replace('.yaml', '')

                    if api_name in ignore:
                        log.info("Ignoring api %s" % api_name)
                        continue

                    apis[api_name] = os.path.join(path, f)
                    log.debug("Found api %s in %s" % (api_name, f))

        # Now generate the model and app code for all these files
        for api_name, api_path in apis.items():
            app_pkg = apipool.load_swagger(
                api_name,
                api_path,
                dest_dir=os.path.dirname(api_path),
                create_endpoints=False if api_name in only_models else True,
                force=force,
                # TODO: custom formats
                # formats=self.formats,
            )
            if app_pkg:
                self.app_pkgs.append(app_pkg)

        if global_names:
            apipool.enforce_global_model_names()

        return self


    def start(self):
        """Load all apis, either as local apis served by the flask app, or as
        remote apis to be called from whithin the app's endpoints, then start
        the app server"""

        self.app.secret_key = os.urandom(24)

        set_error_reporter(self.error_reporter)

        # Initialize JWT config
        conf = get_config()
        if hasattr(conf, 'jwt_secret'):
            log.info("Set JWT parameters to issuer=%s audience=%s secret=%s***" % (
                conf.jwt_issuer,
                conf.jwt_audience,
                conf.jwt_secret[0:8],
            ))

        # Add ping hooks if any
        if self.ping_hook:
            add_ping_hook(self.ping_hook)

        self.load_builtin_apis()

        # Let's compress returned data when possible
        compress = Compress()
        compress.init_app(self.app)

        # Now execute Flask code declaring API routes
        for app_pkg in self.app_pkgs:
            app_pkg.load_endpoints(
                app=self.app,
                error_callback=self.error_callback,
                # formats=self.formats,
                # local=False,
            )

        log.debug("Argv is [%s]" % '  '.join(sys.argv))
        if 'celery' in sys.argv[0].lower():
            # This code is loading in a celery worker - Don't start the actual flask app.
            log.info("Running in a Celery worker - Not starting the Flask app")
            return

        # Initialize monitoring, if any is defined
        monitor_init(app=self.app, config=conf)

        if os.path.basename(sys.argv[0]) == 'gunicorn':
            # Gunicorn takes care of spawning workers
            log.info("Running in Gunicorn - Not starting the Flask app")
            return

        # Debug mode is the default when not running via gunicorn
        self.app.debug = self.debug

        self.app.run(host='0.0.0.0', port=self.port)

#
# Generic code to start server, from command line or via gunicorn
#


def show_splash():
    log.info("")
    log.info("")
    log.info("")
    log.info("       _ __  _   _ _ __ ___   __ _  ___ __ _ _ __ ___  _ __ ")
    log.info("      | '_ \| | | | '_ ` _ \ / _` |/ __/ _` | '__/ _ \| '_ \ ")
    log.info("      | |_) | |_| | | | | | | (_| | (_| (_| | | | (_) | | | |")
    log.info("      | .__/ \__, |_| |_| |_|\__,_|\___\__,_|_|  \___/|_| |_|")
    log.info("      | |     __/ |")
    log.info("      |_|    |___/")
    log.info("")
    log.info("       microservices made easy    -     http://pymacaron.com")
    log.info("")
    log.info("")
    log.info("")


def letsgo(name, callback=None):
    assert callback

    @click.command()
    @click.option('--port', help="Set server listening port (default: 80)", default=None)
    @click.option('--env', help="Set the environment, hence forcing to run against live, staging or dev by setting the PYM_ENV variable", default=None)
    @click.option('--debug/--no-debug', default=True)
    def main(port, env, debug):

        if env:
            log.info("Overriding PYM_ENV to '%s'" % env)
            os.environ['PYM_ENV'] = env

        conf = get_config()

        show_splash()
        if not port:
            port = get_port()

        # Start celeryd and redis?
        if hasattr(conf, 'with_async') and conf.with_async:
            from pymacaron_async import start_celery
            start_celery(port, debug)

        # Proceed to start the API server
        callback(port, debug)

    if name == "__main__":
        main()

    if os.path.basename(sys.argv[0]) == 'gunicorn':
        show_splash()
        port = get_port()
        callback(port)



from pymacaron.log import pymlogger
import json
import uuid
import os
import inspect
import sys
import traceback
from flask import request
from pymacaron.utils import get_container_version
from pymacaron.utils import get_app_name
from pymacaron.utils import is_ec2_instance
from pymacaron.config import get_config
from pymacaron.exceptions import PyMacaronException


log = pymlogger(__name__)


try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


def function_name(f):
    return "%s.%s" % (inspect.getmodule(f).__name__, f.__name__)


#
# Default error reporting
#

def default_error_reporter(title=None, data=None, exception=None):
    """By default, error messages are just logged"""
    log.error(f"error: {title}")
    log.error(f"exception: {exception}")
    log.error(f"details:\n{json.dumps(data, indent=4, sort_keys=True)}")


error_reporter = default_error_reporter


def set_error_reporter(f):
    global error_reporter
    error_reporter = f


def do_report_error(title=None, data=None, exception=None):
    global error_reporter
    log.info("Reporting error...")

    error_reporter(
        title=title,
        data=data,
    )

    try:
        pass
    except Exception as e:
        # Don't block on replying to api caller
        log.error(f"An error occured while trying to report this error: {e}\n")
        exc_type, exc_value, exc_traceback = sys.exc_info()
        trace = traceback.format_exception(exc_type, exc_value, exc_traceback, 30)
        log.error(trace)


def postmortem(f=None, t0=None, t1=None, exception=None, args=[], kwargs={}):
    """Print the error's trace, and call the error reporter with a bunch of data on what happened"""

    data = {}

    # Gather data about the exception that occured
    exc_type, exc_value, exc_traceback = sys.exc_info()
    trace = traceback.format_exception(exc_type, exc_value, exc_traceback, 30)

    status = 500
    if isinstance(exception, PyMacaronException):
        status = exception.status

    if status < 500:
        return

    str_trace = '\n'.join(trace)
    log.error(f"ERROR - ERROR - ERROR - ERROR - ERROR - ERROR:\n{str_trace}")

    data.update({
        'trace': trace,

        # Set only on the original error, not on forwarded ones, not on
        # success responses
        'error_id': str(uuid.uuid4()),

        'is_fatal_error': True if status >= 500 else False,

        # Call results
        'time': {
            'start': t0.isoformat(),
            'end': t1.isoformat(),
            'microsecs': (t1.timestamp() - t0.timestamp()) * 1000000,
        },

        # Response details
        'response': {
            'status': status,
            'error_code': exception.code if hasattr(exception, 'code') else 'UNKNOWN',
            'error_description': str(exception),
            'user_message': exception.user_message if hasattr(exception, 'user_message') else None,
        },
    })

    populate_error_report(data)

    fname = function_name(f)
    do_report_error(
        title=f"{fname}(): {exception}",
        data=data,
        exception=exception,
    )


def report_warning(title=None, data={}, exception=None):
    populate_error_report(data)
    do_report_error(
        title=title,
        data=data,
    )


def populate_error_report(data):
    """Add generic stats to the error report"""

    # Are we in aws?
    data['is_ec2_instance'] = is_ec2_instance()

    # If user is authenticated, get her id
    user_data = {
        'id': '',
        'is_auth': 0,
        'ip': '',
    }

    if stack.top:
        # We are in a request context
        user_data['ip'] = request.remote_addr

        if 'X-Forwarded-For' in request.headers:
            user_data['forwarded_ip'] = request.headers.get('X-Forwarded-For', '')

        if 'User-Agent' in request.headers:
            user_data['user_agent'] = request.headers.get('User-Agent', '')

    if hasattr(stack.top, 'current_user'):
        user_data['is_auth'] = 1
        user_data['id'] = stack.top.current_user.get('sub', '')
        for k in ('name', 'email', 'is_expert', 'is_admin', 'is_support', 'is_tester', 'language'):
            v = stack.top.current_user.get(k, None)
            if v:
                user_data[k] = v

    data['user'] = user_data

    # Server info
    server = request.base_url
    server = server.replace('http://', '')
    server = server.replace('https://', '')
    server = server.split('/')[0]
    parts = server.split(':')
    fqdn = parts[0]
    port = parts[1] if len(parts) == 2 else ''

    data['server'] = {
        'fqdn': fqdn,
        'port': port,
        'api_name': get_app_name(),
        'api_version': get_container_version(),
        'PYM_ENV': os.environ.get('PYM_ENV', ''),
    }

    # Endpoint data
    body_str = ''
    query_str = ''
    try:
        body_str = str(request.get_data())
        query_str = str(request.get.args)
    except Exception:
        pass

    data['request'] = {
        'id': f"{get_app_name()}, {request.method}, {request.path}",
        'url': request.url,
        'base_url': request.base_url,
        'path': request.path,
        'method': request.method,
        'request_body': body_str,
        'request_query': query_str,
    }


#
# DEPRECATED
#

def report_error(title=None, data={}, caught=None, is_fatal=False):
    """Format a crash report and send it somewhere relevant. There are two
    types of crashes: fatal crashes (backend errors) or non-fatal ones (just
    reporting a glitch, but the api call did not fail)"""

    log.info("Caught error: %s\ndata=%s" % (title, json.dumps(data, indent=4)))

    # Don't report errors if NO_ERROR_REPORTING set to 1 (set by run_acceptance_tests)
    if os.environ.get('DO_REPORT_ERROR', None):
        # Force error reporting
        pass
    elif os.environ.get('NO_ERROR_REPORTING', '') == '1':
        log.info("NO_ERROR_REPORTING is set: not reporting error!")
        return
    elif 'is_ec2_instance' in data:
        if not data['is_ec2_instance']:
            # Not running on amazon: no reporting
            log.info("DATA[is_ec2_instance] is False: not reporting error!")
            return
    elif not is_ec2_instance():
        log.info("Not running on an EC2 instance: not reporting error!")
        return

    # Fill error report with tons of usefull data
    if 'user' not in data:
        populate_error_report(data)

    # Add the message
    data['title'] = title
    data['is_fatal_error'] = is_fatal

    # Add the error caught, if any:
    if caught:
        data['error_caught'] = "%s" % caught

    # Add a trace - Formatting traceback may raise a UnicodeDecodeError...
    data['stack'] = []
    try:
        data['stack'] = [l for l in traceback.format_stack()]
    except Exception:
        data['stack'] = 'Skipped trace - contained non-ascii chars'

    # inspect may raise a UnicodeDecodeError...
    fname = ''
    try:
        fname = inspect.stack()[1][3]
    except Exception as e:
        fname = 'unknown-method'

    # Format the error's title
    status, code = 'unknown_status', 'unknown_error_code'
    app_name = get_config().name
    if 'response' in data:
        status = data['response'].get('status', status)
        code = data['response'].get('error_code', code)
        title_details = "%s %s %s" % (app_name, status, code)
    else:
        title_details = "%s %s()" % (app_name, fname)

    if is_fatal:
        title_details = 'FATAL ERROR %s' % title_details
    else:
        title_details = 'NON-FATAL ERROR %s' % title_details

    if title:
        title = "%s: %s" % (title_details, title)
    else:
        title = title_details

    global error_reporter
    log.info("Reporting crash...")

    try:
        error_reporter(title, json.dumps(data, sort_keys=True, indent=4))
    except Exception as e:
        # Don't block on replying to api caller
        log.error("Failed to send email report: %s" % str(e))



#
# Generic crash-handler as a decorator
#

def crash_handler(f):
    """Decorate method with pymacaron's generic crash handler"""
    return generate_crash_handler_decorator(None)(f)



import os
import json
from flask import request, jsonify
from werkzeug import FileStorage
from werkzeug.exceptions import ClientDisconnected
from pydantic.error_wrappers import ValidationError
from pymacaron.log import pymlogger
from pymacaron.utils import timenow
from pymacaron import apipool
from pymacaron import jsonencoders
from pymacaron.model import PymacaronBaseModel
from pymacaron.crash import postmortem
from pymacaron.exceptions import PyMacaronException
from pymacaron.exceptions import UnhandledServerError
from pymacaron.exceptions import InvalidParameterError
from pymacaron.exceptions import BadResponseException
from pymacaron.exceptions import InternalValidationError
from pymacaron.exceptions import RequestTimeout


log = pymlogger(__name__)


def get_form_data(form_args=None):
    # Just extract whatever we can from that form, data or file alike

    try:
        kwargs = request.form.to_dict()
    except ClientDisconnected:
        raise RequestTimeout()

    # Go through all the objects passed in form-data and try converting to something json-friendly
    files = request.files.to_dict()
    for k in list(files.keys()):
        v = files[k]
        if isinstance(v, FileStorage):
            name = v.name
            kwargs[name] = v.read()
        else:
            raise Exception("Support for multipart/form-data containing %s is not implemented" % type(v))

    # If the swagger specified an exact list of form arguments, remove anything
    # that does not match it
    if form_args:
        for k in list(kwargs.keys()):
            if k not in form_args:
                del kwargs[k]

    return kwargs


def get_request_body(api_name, model_name):
    """Return an instantiated pymacaron model containing the request's body or form data"""

    kwargs = {}

    # If the request contained no data, no need to analyze it further
    if len(request.get_data()) != 0:

        # Let's try to convert whatever content-type we got in the request to something json-like
        ctype = request.content_type
        if not ctype:
            # If no content-type specified, assume json
            ctype = 'application/json'

        if ctype.startswith('application/x-www-form-urlencoded'):
            # Get the dict() containing the form's key-values
            kwargs = get_form_data()

        elif ctype.startswith('multipart/form-data'):
            # Store the request's form and files
            kwargs = get_form_data()

        else:
            # Assuming we got a json body
            kwargs = request.get_json(force=True)

    # The pymacaron model class we should instantiate - Let pydantic do all the
    # type checking
    cls = getattr(apipool.get_model(api_name), model_name)

    return cls(**kwargs)


def get_path_and_query_parameters(query_model, path_args):
    # Parse and validate the query arguments (if any)
    d = {}
    if query_model:
        # Let pydantic validate query arguments
        d = query_model.parse_obj(request.args.to_dict()).dict()

    # And add the path arguments (without type checking them: we rely on Flask
    # to have done it)
    d.update(path_args)

    return d


def pymacaron_flask_endpoint(api_name=None, f=None, error_callback=None, query_model=None, body_model_name=None, form_args={}, path_args={}, produces='application/json', result_models=[]):
    """Call endpoint in a try/catch loop handling exceptions"""

    endpoint_method = request.method
    endpoint_path = request.path
    t0 = timenow()

    log.info(" ")
    log.info(" ")
    log.info("=> INCOMING REQUEST %s %s -> %s" % (endpoint_method, endpoint_path, f.__name__))
    log.info(" ")
    log.info(" ")

    try:
        return call_f(
            api_name=api_name,
            f=f,
            query_model=query_model,
            body_model_name=body_model_name,
            path_args=path_args,
            form_args=form_args,
            produces=produces,
            result_models=result_models,
        )

    # Catch ALL exceptions
    except (BaseException, Exception) as e:

        log.error(f"Method {f.__name__} raised exception [{str(e)}]")

        if isinstance(e, ValidationError):
            # Convert this pydantic validation error into a pymacaron one
            e = InvalidParameterError(str(e))

        # Report this exception and tons of info about it
        postmortem(
            f=f,
            t0=t0,
            t1=timenow(),
            exception=e,
        )

        # If it's not a PyMacaronException or a child class of it, it's an
        # unhandled error (aka server crash)
        if not isinstance(e, PyMacaronException):
            e = UnhandledServerError(str(e))

        status = e.status

        if error_callback:
            # The error_callback takes the error instance and returns a json
            # dictionary back
            d = error_callback(e)
            log.info(f"Returning API error (status:{status}): {json.dumps(d, indent=4, sort_keys=True)}")
            r = jsonify(d)
            r.status_code = status
            return r

        return e.jsonify()

    finally:
        log.info(" ")
        log.info("<= DONE %s %s -> %s" % (endpoint_method, endpoint_path, f.__name__))
        log.info(" ")
        log.info(" ")


def call_f(api_name=None, f=None, error_callback=None, query_model=None, body_model_name=None, form_args={}, path_args={}, produces='application/json', result_models=[]):
    """A generic flask endpoint that calls a given pymacaron endpoint
    implementation and handle conversion between query/body parameters,
    pymacaron models and flask response. Also handle error handling and
    optional decoration.

    api_name: name of the api to which this endpoint belongs

    f: reference to the method that implements this endpoint in the pymacaron microservice

    path_query_model_name: name of the pymacaron/pydantic model that defines all of this endpoints query and path parameters

    body_model_name: name of the model that defines the HTTP body data expected by this endpoint (None if none)

    """

    if os.environ.get('PYM_DEBUG', None) == '1':
        log.debug("PYM_DEBUG: Request headers are: %s" % dict(request.headers))

    args = []
    if body_model_name:
        args.append(get_request_body(api_name, body_model_name))

    kwargs = get_path_and_query_parameters(query_model, path_args)

    if form_args:
        kwargs.update(get_form_data(form_args))

    if os.environ.get('PYM_DEBUG', None) == '1':
        log.debug("PYM_DEBUG: Request args are: [args: %s] [kwargs: %s]" % (args, kwargs))

    try:
        result = f(*args, **kwargs)
    except ValidationError as e:
        # A pydantic validation error occuring inside the endpoint is actually
        # a fatal crash. We re-raise it but changed its type
        raise InternalValidationError(str(e)) from e

    if produces == 'application/json':
        assert result_models, "BUG: no result models specified"
        str_result_models = ' or '.join([str(m) for m in result_models])

        if not result:
            raise BadResponseException('Nothing to return in response')

        elif isinstance(result, PymacaronBaseModel):
            # Validate that we got the right model as results
            found = False
            for m in result_models:
                if isinstance(result, m):
                    found = True
            if not found:
                raise BadResponseException(f'Expected to return an instance of {str_result_models}, but got a {result}')

            return jsonify(result.to_json(
                exclude_unset=True,
                exclude_none=False,
                keep_nullable=True,
                keep_datetime=False,
                datetime_encoder=jsonencoders.get_datetime_encoder(),
            ))

        elif ".".join([result.__module__, result.__class__.__name__]) == 'flask.wrappers.Response':
            # result is already a flask response
            return result

        else:
            raise BadResponseException(f'Expected to return an instance of {str_result_models} but got {result} of type {type(result)}')

    else:
        # TODO: implement support for returning html content
        #     if type(result) is not tuple:
        #         e = error_callback(PyMacaronCoreException("Method %s should return %s but returned %s" %
        #                                                   (endpoint.handler_server, endpoint.produces, type(result))))
        #         return _responsify(api_spec, e, 500)

        #     # Return an html page
        #     return result

        assert 0, "Support for returning '{produces}' not implemented yet!"



import jwt
from urllib.parse import unquote_plus
from contextlib import contextmanager
from functools import wraps
from flask import request
from pymacaron.log import pymlogger
from pymacaron.exceptions import AuthInvalidTokenError
from pymacaron.exceptions import AuthTokenExpiredError
from pymacaron.exceptions import AuthMissingHeaderError
from pymacaron.utils import timenow, to_epoch
from pymacaron.config import get_config

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


log = pymlogger(__name__)


#
# Decorators used to add authentication to endpoints in swagger specs
#

def requires_auth(f):
    """A decorator for flask api methods that validates auth0 tokens, hence ensuring
    that the user is authenticated. Code coped from:
    https://github.com/auth0/auth0-python/tree/master/examples/flask-api
    """

    @wraps(f)
    def requires_auth_decorator(*args, **kwargs):
        authenticate_http_request()
        return f(*args, **kwargs)

    return requires_auth_decorator


def add_auth(f):
    """A decorator that adds the authentication header to requests arguments"""

    def add_auth_decorator(*args, **kwargs):
        token = get_user_token()
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['Authorization'] = f"Bearer {token}"
        return f(*args, **kwargs)

    return add_auth_decorator


#
# Get and validate a token
#

def load_auth_token(token, load=True):
    """Validate an auth0 token. Returns the token's payload, or an exception
    of the type:"""

    conf = get_config()

    assert conf.jwt_secret, "No JWT secret configured for pymacaron"
    assert conf.jwt_issuer, "No JWT issuer configured for pymacaron"
    assert conf.jwt_audience, "No JWT audience configured for pymacaron"

    # First extract the issuer
    issuer = conf.jwt_issuer
    try:
        headers = jwt.get_unverified_header(token)
    except jwt.exceptions.DecodeError:
        raise AuthInvalidTokenError('token signature is invalid')

    log.info(f"JWT token has headers '{headers}'")

    if 'iss' in headers:
        issuer = headers['iss']

    # Then validate the token against this issuer
    try:
        payload = jwt.decode(
            token,
            conf.jwt_secret,
            audience=conf.jwt_audience,
            algorithms=["HS256"],
            # Allow for a time difference of up to 5min (300sec)
            leeway=300,
        )
    except jwt.exceptions.ExpiredSignatureError:
        log.debug('JWT token has expired')
        raise AuthTokenExpiredError('JWT token has expired')
    except jwt.exceptions.InvalidSignatureError:
        log.debug('JWT token signature is invalid')
        raise AuthInvalidTokenError('JWT token signature is invalid')
    except jwt.exceptions.InvalidAudienceError:
        log.debug('JWT has incorrect audience')
        raise AuthInvalidTokenError('JWT has incorrect audience')
    except jwt.exceptions.InvalidIssuedAtError:
        log.debug('JWT token was issued in the future')
        raise AuthInvalidTokenError('JWT token was issued in the future')
    except jwt.exceptions.InvalidTokenError:
        # A catch-all for all other invalid token errors
        log.debug('JWT token is invalid')
        raise AuthInvalidTokenError('JWT token is invalid')

    # Save payload to stack
    payload['token'] = token
    payload['iss'] = issuer

    if load:
        stack.top.current_user = payload

    return payload

def authenticate_http_request(token=None):
    """Validate auth0 tokens passed in the request's header, hence ensuring
    that the user is authenticated. Code copied from:
    https://github.com/auth0/auth0-python/tree/master/examples/flask-api

    Return a PntCommonException if failed to validate authentication.
    Otherwise, return the token's payload (Also stored in stack.top.current_user)
    """

    if token:
        auth = token
    else:
        auth = request.headers.get('Authorization', None)

    if not auth:
        auth = request.cookies.get('token', None)
        if auth:
            auth = unquote_plus(auth)

    log.debug(f"Validating Auth header [{auth}]")

    if not auth:
        raise AuthMissingHeaderError('There is no Authorization header in the HTTP request')

    parts = auth.split()

    if parts[0].lower() != 'bearer':
        raise AuthInvalidTokenError('Authorization header must start with Bearer')
    elif len(parts) == 1:
        raise AuthInvalidTokenError('Token not found in Authorization header')
    elif len(parts) > 2:
        raise AuthInvalidTokenError('Authorization header must be Bearer + \\s + token')

    token = parts[1]

    return load_auth_token(token)

#
# Generate tokens
#

def generate_token(user_id, expire_in=None, data={}, issuer=None, iat=None):
    """Generate a new JWT token for this user_id. Default expiration date
    is 1 year from creation time"""
    assert user_id, "No user_id passed to generate_token()"
    assert isinstance(data, dict), "generate_token(data=) should be a dictionary"
    assert get_config().jwt_secret, "No JWT secret configured in pymacaron"

    if not issuer:
        issuer = get_config().jwt_issuer

    assert issuer, "No JWT issuer configured for pymacaron"

    if expire_in is None:
        expire_in = get_config().jwt_token_timeout

    if iat:
        epoch_now = iat
    else:
        epoch_now = to_epoch(timenow())
    epoch_end = epoch_now + expire_in

    data['iss'] = issuer
    data['sub'] = user_id
    data['aud'] = get_config().jwt_audience
    data['exp'] = epoch_end
    data['iat'] = epoch_now

    headers = {
        "typ": "JWT",
        "alg": "HS256",
        "iss": issuer,
    }

    log.debug("Encoding token with data '{data}' and headers '{headers}'")

    t = jwt.encode(
        data,
        get_config().jwt_secret,
        headers=headers,
    )

    if type(t) is bytes:
        t = t.decode("utf-8")

    return t


@contextmanager
def backend_token(issuer=None, user_id=None, data={}):

    if not issuer:
        issuer = get_config().jwt_issuer
    if not user_id:
        user_id = get_config().default_user_id

    assert issuer, "No JWT issuer configured for pymacaron"
    assert user_id, "No user_id passed to generate_token()"

    cur_token = ''

    if stack.top is None:
        raise RuntimeError('working outside of request context')

    if not hasattr(stack.top, 'current_user'):
        stack.top.current_user = {}
    else:
        cur_token = stack.top.current_user.get('token', '')

    tmp_token = generate_token(user_id, issuer=issuer, data=data)

    log.debug("Temporarily using custom token for %s and issuer %s: %s" % (user_id, issuer, tmp_token))
    stack.top.current_user['token'] = tmp_token
    yield tmp_token
    log.debug("Restoring token %s" % cur_token)
    stack.top.current_user['token'] = cur_token


#
# Access token data during runtime
#


def get_userid():
    """Return the authenticated user's id, i.e. its auth0 id"""
    try:
        current_user = stack.top.current_user
        return current_user.get('sub', '')
    except Exception:
        return ''


def get_user_token_data():
    """Return the authenticated user's id, i.e. its auth0 id"""
    return stack.top.current_user


def get_user_token():
    """Return the authenticated user's auth token"""
    if not hasattr(stack.top, 'current_user'):
        return ''
    current_user = stack.top.current_user
    return current_user.get('token', '')


def get_token_issuer():
    """Return the issuer in which this user's token was created"""
    try:
        current_user = stack.top.current_user
        return current_user.get('iss', get_config().jwt_issuer)
    except Exception:
        pass
    return get_config().jwt_issuer



import os
import sys
from pymacaron.log import pymlogger
import datetime
from pymacaron.config import get_config
from dateutil import parser
import socket
import pytz


log = pymlogger(__name__)


IS_EC2_INSTANCE = -1


def is_ec2_instance():
    """Try fetching instance metadata at 'curl http://169.254.169.254/latest/meta-data/'
    to see if host is on an ec2 instance"""

    # Note: this code assumes that docker containers running on ec2 instances
    # inherit instances metadata, which they do as of 2016-08-25

    global IS_EC2_INSTANCE

    if IS_EC2_INSTANCE != -1:
        # Returned the cached value
        return IS_EC2_INSTANCE

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.2)
    try:
        s.connect(("169.254.169.254", 80))
        IS_EC2_INSTANCE = 1
        return True
    except socket.timeout:
        IS_EC2_INSTANCE = 0
        return False
    except socket.error:
        IS_EC2_INSTANCE = 0
        return False


def timenow():
    return datetime.datetime.now(pytz.timezone('utc'))


def to_epoch(t):
    """Take a datetime, either as a string or a datetime.datetime object,
    and return the corresponding epoch"""
    if isinstance(t, str):
        if '+' not in t:
            t = t + '+00:00'
        t = parser.parse(t)
    elif t.tzinfo is None or t.tzinfo.utcoffset(t) is None:
        t = t.replace(tzinfo=pytz.timezone('utc'))

    t0 = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, pytz.timezone('utc'))
    delta = t - t0
    return int(delta.total_seconds())


def to_datetime(e):
    """Take an epoch and return a timezone aware datetime"""
    return datetime.datetime.fromtimestamp(e, pytz.timezone('utc'))


def get_container_version():
    """Return the version of the docker container running the present server,
    or '' if not in a container"""
    version_file = '/pym/VERSION'
    if os.path.exists(version_file):
        with open(version_file) as f:
            return f.read().strip()
    return 'local'


def get_app_name():
    return get_config().name


def __prune_none(j: dict):
    """Remove all keys that are set to None in this python dictionary, recursively"""
    for k in list(j.keys()):
        v = j[k]
        if v is None:
            del j[k]
        elif type(v) is dict:
            __prune_none(v)
        elif type(v) is list:
            for vv in v:
                if type(vv) is dict:
                    __prune_none(vv)


def prune_none(*args, **kwargs):
    """Takes either a python dictionary, or a **kwargs key-value pairs, and remove
    all keys that are set to None, recursively on dictionaries and list of
    dictionaries.

    Usage:

    apipool.myapi.MyModel(**prune_none(foo='bar', bar=None))

    apipool.myapi.MyModel(**prune_none({'foo': 'bar', bar: None, 'baz': {'a': None, 'b': 1}))

    """

    if len(args) == 1:
        j = args[0]
        assert type(j) is dict
    elif len(args) > 1:
        j = dict(args)
    else:
        j = kwargs

    __prune_none(j)

    return j



import os
import re
import yaml
import importlib.util
from pymacaron.log import pymlogger


log = pymlogger(__name__)


def modified_time(path):
    return os.path.getmtime(path)


def swagger_type_to_pydantic_type(t, models=[]):
    """Given a swagger type or format or ref definition, return its corresponding python type"""
    # Take a swagger type and return the corresponding python type
    if t in models:
        return t

    mapping = {
        # swagger type: pydantic type (https://pydantic-docs.helpmanual.io/usage/types)
        'boolean': 'bool',
        'int32': 'int',
        'integer': 'int',
        'string': 'str',
        'number': 'float',
        'date': 'datetime',
        'datetime': 'datetime',
        'date-time': 'datetime',
        'iso-date': 'datetime',
    }

    assert t in mapping, f"Don't know to map swagger type '{t}' to a python type"

    return mapping[t]


def ref_to_model_name(s):
    """Take the '#/definitions/...' string in a $ref: statement, and return the name of the model referenced"""
    return s.split('/')[-1].replace("'", "").replace('"', '').strip()


def generate_models_v2(swagger, model_file, api_name):
    """Extract all model definitions from this swagger file and write a python
    module defining every corresponding pymacaron model class
    """

    log.info(f"Regenerating {model_file}")

    # List all model names in definitions
    all_models = list(swagger['definitions'].keys())
    all_models.sort()
    str_all_models = ', '.join([f'"{s}"' for s in all_models])

    def def_to_type(prop_def, model_name, prop_name):
        """Figure out the python type of a property"""
        assert type(prop_def) is dict, f"Expected a dict instead of '{prop_def}' in definition of {model_name}:{prop_name}"
        t = None
        if '$ref' in prop_def:
            s = prop_def['$ref']
            assert s.startswith("#/definitions/"), f"Failed to parse ref '{s}' in definition of {model_name}:{prop_name}"
            t = ref_to_model_name(s)
        elif 'enum' in prop_def:
            assert type(prop_def['enum']) is list
            str_enum = ', '.join(f'"{s}"' for s in prop_def['enum'])
            t = f'Literal[None, {str_enum}]'
            return t
        elif 'format' in prop_def:
            t = prop_def['format']
        elif 'type' in prop_def:
            t = prop_def['type']
        else:
            raise Exception(f"Don't know how to identify type in '{prop_def}' definition of {model_name}:{prop_name}")

        if t in all_models:
            return t

        return swagger_type_to_pydantic_type(t, all_models)

    def get_parent(model_def):
        """Return (parent_module_path, parent_class) or None for this model definition"""
        parent = model_def.get('x-parent', None)
        if parent:
            cls = parent.split('.')[-1]
            path = '.'.join(parent.split('.')[0:-1])
            return (path, cls)
        return (None, None)

    def def_to_model_dependencies(model_def, model_name):
        """Given a model definition, return all the model it references to"""
        deps = []
        assert 'properties' in model_def, f"No 'properties' defined for model '{model_name}'"
        for prop_name, prop_def in model_def['properties'].items():
            if prop_def.get('type', '').lower() == 'array':
                assert 'items' in prop_def, f"Expected 'items' in array definition of {model_name}:{prop_name}"
                t = def_to_type(prop_def['items'], model_name, prop_name)
            else:
                t = def_to_type(prop_def, model_name, prop_name)
            if t in all_models:
                deps.append(t)
        return deps

    # List all the parent classe we need to import before declaring models
    imports = []
    for model_name, model_def in swagger['definitions'].items():
        (path, cls) = get_parent(model_def)
        if path and cls:
            imports.append(f'from {path} import {cls} as Parent{cls}')

    # Code lines of the python file
    lines = [
        '# This is an auto-generated file - DO NOT EDIT!!!',
        'from pymacaron.model import PymacaronBaseModel',
        'from pydantic import BaseModel',
        'from typing import List, Optional, Literal',
        'from datetime import datetime',
    ] + imports + [
        '',
        f'__all_models = [{str_all_models}]',
        '',
    ]

    # Now comes a tricky part: we need to sort class declarations so that
    # classes are declared before they are referenced in the type constraints
    # of other classes.

    ordered_names = []
    deps_by_name = {}
    for model_name, model_def in swagger['definitions'].items():
        deps_by_name[model_name] = def_to_model_dependencies(model_def, model_name)

    while len(deps_by_name):
        for name in list(deps_by_name.keys()):
            all_declared = True
            for dep in deps_by_name[name]:
                if dep not in ordered_names:
                    all_declared = False
            if all_declared:
                ordered_names.append(name)
                del deps_by_name[name]

    # Now let's declare every class, in the right order
    for model_name in ordered_names:
        model_def = swagger['definitions'][model_name]

        str_properties = ', '.join([f'"{s}"' for s in list(model_def['properties'].keys())])

        str_nullable_properties = ', '.join([f'"{s}"' for s, d in model_def['properties'].items() if d.get('x-nullable', False) is True])

        (path, cls) = get_parent(model_def)
        x_parent = f'Parent{cls}, ' if cls else ''

        lines += [
            '',
            '',
            f'class {model_name}({x_parent}PymacaronBaseModel, BaseModel):',
            '    def get_property_names(self):',
            f'        return [{str_properties}]',
            '    def get_model_api(self):',
            f'        return "{api_name}"',
            '    def get_nullable_properties(self):',
            f'        return [{str_nullable_properties}]',
        ]

        for p_name, p_def in model_def['properties'].items():

            p_type = ''
            if p_def.get('type', '').lower() == 'array':
                assert 'items' in p_def, f"Expected 'items' in array definition of {model_name}:{p_name}"
                t = def_to_type(p_def['items'], model_name, p_name)
                p_type = f'List[{t}]'
            else:
                t = def_to_type(p_def, model_name, p_name)
                p_type = t

            # By default, pydantic attributes are optional. But if the property
            # has 'x-mandatory: true', then it becomes required.
            if p_def.get('x-mandatory', False) is True:
                lines.append(f'    {p_name}: {p_type}')
            else:
                lines.append(f'    {p_name}: Optional[{p_type}] = None')

    lines.append('')
    lines.append('')

    with open(model_file, 'w') as f:
        f.write('\n'.join(lines))


def generate_endpoints_v2(swagger, app_file, model_file, api_name):
    """Extract all endpoint definitions from this swagger file and write a python
    module defining all the corresponding FastAPI endpoints
    """

    def swagger_type_to_flask_converter(s):
        mappings = {
            # swagger -> flask
            'string': 'string',
            'integer': 'int',
            'int32': 'int',
            'number': 'float',
        }
        assert s in mappings, f"Don't know how to map swagger type '{s}' of a path parameter to a Flask converter"
        return mappings[s]

    if 'paths' not in swagger:
        log.info(f"No routes defined in {app_file}")
        return

    log.info(f"Regenerating {app_file}")

    lines_header = [
        '# This is an auto-generated file - DO NOT EDIT!!!',
        'from flask_cors import cross_origin',
        'from typing import Optional',
        'from pydantic import BaseModel',
        'from pymacaron.endpoint import pymacaron_flask_endpoint',
        'from pymacaron.log import pymlogger',
        '',
        '',
        'log = pymlogger(__name__)',
        '',
        '',
        'def load_endpoints(app=None, error_callback=None):',
        '    from pymacaron import apipool',
        '',
    ]

    lines_imports = []

    lines_endpoints = [
    ]

    for route in swagger['paths'].keys():
        for http_method, endpoint_def in swagger['paths'][route].items():

            http_method = http_method.upper()

            # Extract path, query and body parameters from in declaration
            query_params = {}
            form_params = {}
            path_params = {}
            str_body_model_name = 'None'

            if 'parameters' in endpoint_def:
                for param in endpoint_def['parameters']:
                    err_str = f'{param} of route {http_method}:{route} in api "{api_name}"'
                    assert 'in' in param, f"Missing 'in' declaration {err_str}"
                    if param['in'] == 'path':
                        assert 'name' in param and 'type' in param, f"Missing 'name' and/or 'type' declaration for path paramater {err_str}"
                        path_params[param['name']] = param['type']
                    elif param['in'] == 'query':
                        assert 'name' in param and 'type' in param, f"Missing 'name' and/or 'type' declaration for query paramater {err_str}"
                        query_params[param['name']] = param['type']
                    elif param['in'] == 'body':
                        assert 'schema' in param, f"Missing 'schema' declaration for body parameter {err_str}"
                        assert '$ref' in param['schema'], f"Missing '$ref' declaration in schema of body parameter {err_str}"
                        s = ref_to_model_name(param['schema']['$ref'])
                        str_body_model_name = f'"{s}"'
                    elif param['in'] == 'formData':
                        assert 'name' in param, f"Missing 'name' declaration for formData parameter {err_str}"
                        assert 'type' in param, f"Missing 'type' declaration for formData parameter {err_str}"
                        assert param['type'] == 'file', f"Type {param['type']} is not supported for formData in parameter {err_str}"
                        form_params[param['name']] = param['type']
                    else:
                        assert 0, f"param type {param['in']} not supported yet in parameter {err_str}"

            # What does the endpoint produces?
            assert 'produces' in endpoint_def, f"Missing 'produces' in declaration of endpoint {http_method}:{route} in api '{api_name}'"
            produces = endpoint_def['produces'][0]
            assert produces == 'application/json'

            # Model returned?
            assert 'responses' in endpoint_def, f"Missing 'responses' in declaration of endpoint {http_method}:{route} in api '{api_name}'"
            result_models_lines = []
            for response_type, response_def in endpoint_def['responses'].items():
                assert 'schema' in response_def, f"Missing 'schema' in response '{response_type}' in declaration of endpoint {http_method}:{route} in api '{api_name}'"
                assert '$ref' in response_def['schema'], f"Missing '$ref' in response '{response_type}' in declaration of endpoint {http_method}:{route} in api '{api_name}'"
                s = ref_to_model_name(response_def['schema']['$ref'])
                result_models_lines += [
                    f'                apipool.{api_name}.{s},',
                ]

            # Extract x-bind-server
            operation_id = None
            if 'operationId' in endpoint_def:
                operation_id = endpoint_def['operationId']
            elif 'x-bind-server' in endpoint_def:
                operation_id = endpoint_def['x-bind-server']
            else:
                raise Exception(f"Endpoint {http_method}:{route} has no 'x-bind-server' or 'operationId' declaration")
            method_path = '.'.join(operation_id.split('.')[0:-1])
            method_name = operation_id.split('.')[-1]
            unique_method_name = 'f_' + operation_id.replace('.', '_')

            # Name the endpoint
            def_name = f'endpoint_{http_method.lower()}_' + re.sub(r'\W+', '', str(route).replace('/', '_'))

            # Import the method that implements the endpoint and apply its decorator, if any
            if 'x-decorate-server' in endpoint_def:
                s = endpoint_def['x-decorate-server']
                decorator_f = s.split('.')[-1]
                decorator_pkg = '.'.join(s.split('.')[0:-1])

                # Rename unique_method_name to indicate that it is decorated
                # (so as to not conflict with the same method imported without
                # decorator)
                unique_method_name = f'{unique_method_name}_via_{decorator_f}'

                lines_imports += [
                    '',
                    f'    from {method_path} import {method_name} as {unique_method_name}',
                    f'    from {decorator_pkg} import {decorator_f}',
                    f'    {unique_method_name} = {decorator_f}({unique_method_name})',
                ]
            else:
                lines_imports += [
                    '',
                    f'    from {method_path} import {method_name} as {unique_method_name}',
                ]

            # Compile flask route from swagger route (s/{}/<>/ + add :type)
            flask_route = str(route)
            for name, typ in path_params.items():
                flask_converter = swagger_type_to_flask_converter(typ)
                flask_route = flask_route.replace('{' + name + '}', f'<{flask_converter}:{name}>')

            # If there are query parameters, generate a pydantic model used to
            # parse their string representation
            str_query_model = 'None'
            query_model_lines = []
            if len(query_params):
                str_query_model = 'QueryModel'
                query_model_lines = [
                    '        class QueryModel(BaseModel):',
                ]
                for name, typ in query_params.items():
                    python_type = swagger_type_to_pydantic_type(typ)
                    query_model_lines += [
                        f'            {name}: Optional[{python_type}] = None',
                    ]

            # If there are path parameters, pass them as a dictionary to the
            # generic pymacaron endpoint
            path_args_lines = []
            str_path_params = ''
            if len(path_params):
                str_path_params = ', '.join(path_params.keys())
                for name in path_params.keys():
                    path_args_lines += [
                        f'                "{name}": {name},',
                    ]


            # If there are path parameters, pass them as a dictionary to the
            # generic pymacaron endpoint
            form_args_lines = []
            if len(form_params):
                for name, typ in form_params.items():
                    form_args_lines += [
                        f'                "{name}": "{typ}",',
                    ]

            lines_endpoints += [
                '',
                f'    @app.route("{flask_route}", methods=["{http_method}"])',
                '    @cross_origin(headers=["Content-Type", "Authorization"])',
                f'    def {def_name}({str_path_params}):',

            ] + query_model_lines + [
                '        return pymacaron_flask_endpoint(',
                f'            api_name="{api_name}",',
                f'            f={unique_method_name},',
                '            path_args={',
            ] + path_args_lines + [
                '            },',
                '            form_args={',
            ] + form_args_lines + [
                '            },',
                f'            body_model_name={str_body_model_name},',
                f'            query_model={str_query_model},',
                f'            produces="{produces}",',
                '            result_models=[',
            ] + result_models_lines + [
                '            ],',
                '            error_callback=error_callback,',
                '        )',
                f'    log.info("Binding [{api_name}] {http_method} {route} ==> {operation_id}")',
                '',
            ]

    # log.info("GET /api/v4/chat/<chat_id>/message/<message_id> ==> gofrendly.v4.message.do_get_message")
    # @app.route('/api/v4/chat/<str:chat_id>/message/<str:message_id>', methods=['GET'])
    # def endpoint_do_get_message(chat_id, message_id):
    #     from gofrendly.v4.message import do_get_message
    #     return pymacaron_flask_endpoint(
    #         api_name='chat',
    #         f=do_get_message,
    #         path_args={
    #             'chat_id': chat_id,
    #             'message_id': message_id,
    #         },
    #         body_model_name=None,
    #         query_model=None,
    #     )

    # log.info("GET /api/v4/chat/<chat_id>/message/add ==> gofrendly.v4.message.do_post_message")
    # @app.route('/api/v4/chat/<str:chat_id>/message/add', methods=['POST'])
    # def endpoint_do_post_message(chat_id):
    #     from gofrendly.v4.message import do_post_message
    #     return pymacaron_flask_endpoint(
    #         api_name='chat',
    #         f=do_post_message,
    #         path_args={
    #             'chat_id': chat_id,
    #         },
    #         body_model_name='v4NewChatMessage',
    #         query_model=None,
    #     )

    # log.info("GET /api/v4/chat/<chat_id>/messages/search ==> gofrendly.v4.message.do_search_messages")
    # @app.route('/api/v4/chat/<str:chat_id>/messages/search', methods=['GET'])
    # class endpoint_do_search_messages(BaseModel):
    #     class QueryModel(BaseModel):
    #         text: Optional[str] = None
    #         id: Optional[str] = None
    #
    #     from gofrendly.v4.message import do_search_messages
    #     return pymacaron_flask_endpoint(
    #         api_name='chat',
    #         f=do_search_message,
    #         path_args={
    #             'chat_id': chat_id,
    #         },
    #         body_model_name=None,
    #         query_model=QueryModel,
    #     )

    # Optionally: global decorator
    # log.info(...)
    # @app.route('/api/v4/chat/<str:chat_id>/messages/search', methods=['GET'])
    # @global_decorator()
    # class...

    with open(app_file, 'w') as f:
        f.write('\n'.join(lines_header + lines_imports + lines_endpoints))


def load_api_models_and_endpoints(api_name=None, api_path=None, dest_dir=None, model_file=None, app_file=None, create_endpoints=True, force=False, load_code=True):
    """Load all object models defined inside the OpenAPI specification located at
    api_path into a generated python module at dest_dir/[api_name].py

    Optionally generate Flask endpoints and route declarations, binding the
    APIs routes to the methods implementing them, and converting request
    parameters and response to and from pymacaron models.

    If endpoints were generated, return a module object with a load_endpoints()
    method to call with the Flask app in argument.

    """

    if not dest_dir:
        dest_dir = os.path.dirname(api_path)

    if model_file is None:
        model_file = './' + os.path.relpath(os.path.join(dest_dir, f'{api_name}_models.py'))
    if app_file is None:
        app_file = './' + os.path.relpath(os.path.join(dest_dir, f'{api_name}_app.py'))

    #
    # Step 1: Regenerate pydantic and FastAPI python code, if needed
    #

    # Should we re-generate the models file?
    do_models = False
    if force:
        do_models = True
    elif not os.path.exists(model_file):
        do_models = True
    elif modified_time(model_file) < modified_time(api_path):
        do_models = True

    # Should we re-generate the endpoints file?
    do_endpoints = False
    if create_endpoints:
        if force:
            do_endpoints = True
        elif not os.path.exists(model_file):
            do_endpoints = True
        elif modified_time(model_file) < modified_time(api_path):
            do_endpoints = True

    # Do we need to re-generate anything?
    if do_models or do_endpoints:
        swagger = None

        if api_path.endswith('.yaml'):
            log.info(f"Loading swagger file {api_path}")
            swagger = yaml.load(open(api_path), Loader=yaml.FullLoader)

        if not swagger:
            raise Exception(f"Don't know how to load {api_path}")

        swagger_version = swagger['swagger']
        assert swagger_version == '2.0', f"OpenAPI version '{swagger_version}' is not supported"

        if do_models:
            generate_models_v2(swagger, model_file, api_name)

        if do_endpoints:
            generate_endpoints_v2(swagger, app_file, model_file, api_name)

    else:
        if not do_models:
            log.info(f"No need to regenerate {model_file}")
        if create_endpoints and not do_endpoints:
            log.info(f"No need to regenerate {app_file}")

    #
    # Step 2: Load code
    #

    def do_load_code(path):
        spec = importlib.util.spec_from_file_location(api_name, path)
        pkg = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(pkg)
        return pkg

    model_pkg = None
    if load_code:
        model_pkg = do_load_code(model_file)

    app_pkg = None
    if create_endpoints:
        app_pkg = do_load_code(app_file)

    #
    # Step 3: Load all pydantic models into apipool
    #

    from pymacaron import apipool

    if model_pkg:
        cnt = 0
        for model_name in model_pkg.__all_models:
            apipool.add_model(api_name, model_name, getattr(model_pkg, model_name))
            cnt += 1

        log.info(f"Loaded {cnt} models from {model_file}")

    return app_pkg



import ujson
from difflib import unified_diff
from datetime import datetime
from pymacaron.log import pymlogger
from pymacaron.utils import prune_none as do_prune_none


log = pymlogger(__name__)


class PymacaronBaseModel(object):
    """The base class from which all pymacaron model classes inherit. Some of these
    methods are redundant with pydantic, but kept for backward compatibility
    with code using older versions of pymacaron.

    """

    # See https://pydantic-docs.helpmanual.io/usage/exporting_models/ about ujson vs orjson
    class Config:
        # Do type validation each time a property is set
        validate_assignment = True

        # Don't allow extra properties when passing kwargs to __init__()
        extra = 'forbid'

        # # extra json.dumps arguments
        # indent = 4
        # sort_keys = True


    def __str__(self):
        """Return a generic string representation of a pymacaron model instance"""
        return f'{self.get_model_name()}(self.dict())'


    def __set_nullable(self, j, o):
        # Set x-nullable keys to None if they are missing
        for k in o.get_nullable_properties():
            if k not in j:
                j[k] = None
        for k in list(j.keys()):
            v = j[k]
            if type(v) is dict:
                self.__set_nullable(v, getattr(o, k))
            elif type(v) is list:
                for i in range(len(v)):
                    jj = v[i]
                    if type(jj) is dict:
                        self.__set_nullable(jj, getattr(o, k)[i])


    def __serialize_datetime(self, j, f):
        for k, v in j.items():
            if type(v) is datetime:
                j[k] = f(v)
            elif type(v) is dict:
                self.__serialize_datetime(v, f)
            elif type(v) is list:
                for jj in v:
                    if type(jj) is dict:
                        self.__serialize_datetime(jj, f)


    def to_json(self, keep_datetime=False, keep_nullable=False, datetime_encoder=None, exclude_unset=True, exclude_none=False):
        """Return a dictionary representation this PyMacaron object, with datetime
        types serialized to strings.
        """

        # We want a json dict of only the object's properties that have been set.
        # https://stackoverflow.com/questions/66229384/pydantic-detect-if-a-field-value-is-missing-or-given-as-null
        j = self.dict(exclude_unset=exclude_unset, exclude_none=exclude_none)

        # Optionally set nullable fields to None
        if keep_nullable:
            self.__set_nullable(j, self)

        # Optionally serialize datetimes
        if not keep_datetime:
            if not datetime_encoder:
                # The default encoding of datetime is .isoformat()
                self.__serialize_datetime(j, lambda d: d.isoformat())
            else:
                self.__serialize_datetime(j, datetime_encoder)

        return j


    @classmethod
    def from_json(cls, j, prune_none=True):
        """Take a json dictionary and return a model instance"""
        if prune_none:
            do_prune_none(j)
        return cls.parse_obj(j)


    def clone(self):
        # Deprecated: should use pydantic.copy() instead
        return self.copy(deep=True)


    def get_model_name(self):
        """Return the name of the OpenAPI schema object describing this pymacaron Model instance"""
        return type(self).__name__


    def get_model_api(self):
        """Return the name of the api to which this model belongs"""
        raise Exception("Should be overriden in model declaration")


    def get_property_names(self):
        """Return the names of all of the model's properties"""
        raise Exception("Should be overriden in model declaration")


    @classmethod
    def diff_with(me, other):
        """Return a diff-style text showing the difference between the
        PymacaronBaseModels schemas of self and other. Used to see if 2
        pymacaron objects of different types have the same schema definition.
        """

        d0 = me.schema()
        d1 = other.schema()

        # strip all descriptions
        def __remove_descriptions(d):
            for k in list(d.keys()):
                v = d[k]
                if k == 'description':
                    del d[k]
                elif type(v) is dict:
                    __remove_descriptions(v)
                elif type(v) is list:
                    for dd in v:
                        if type(dd) is dict:
                            __remove_descriptions(dd)

        __remove_descriptions(d0)
        __remove_descriptions(d1)

        s0 = ujson.dumps(d0, indent=4, sort_keys=True)
        s1 = ujson.dumps(d1, indent=4, sort_keys=True)

        generator = unified_diff(
            s0.splitlines(keepends=True),
            s1.splitlines(keepends=True),
            fromfile=f'{me.get_model_api(me)}.{me.__name__}',
            tofile=f'{other.get_model_api(other)}.{other.__name__}',
            lineterm='\n',
        )

        return ''.join(generator).strip()
